import argparse
import json
import logging
import os
from typing import Any, Callable, Dict, List, Mapping, TypedDict, Union, cast

from mergedeep import merge
from mitogen.core import Error, Receiver, StreamError
from mitogen.parent import Context, EofError, Router
from mitogen.ssh import HostKeyError
from mitogen.utils import run_with_router
from retry.api import retry_call

from paracrine import DRY_RUN_ENV

from .deps import (
    Module,
    Modules,
    TransmitModule,
    TransmitModules,
    freeze_module,
    makereal,
    maketransmit,
    maketransmit_single,
    runfunc,
    unfreeze_module,
)
from .helpers.config import (
    ServerDict,
    create_data,
    get_config,
    other_config_file,
    path_to_config_file,
    set_config,
    set_data,
)
from .helpers.fs import set_file_contents
from .runners import core


def decode(info: Dict[Union[bytes, str], Union[bytes, str]]):
    for k in list(info.keys()):
        if isinstance(k, bytes):
            value = info[k]
            assert isinstance(value, bytes)
            info[k.decode()] = value.decode()
            del info[k]


ssh_cache: Dict[str, Context] = {}


def clear_ssh_cache():
    global ssh_cache
    ssh_cache = {}


class MainReturn(TypedDict):
    infos: List[Any]
    data: Mapping[str, object]


def main(
    router: Router,
    servers: Union[list[str], None],
    func: Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> MainReturn:
    config = get_config()
    calls: List[Receiver] = []
    wg = core.is_wireguard()
    data = {}
    for server in config["servers"]:
        if servers is not None and server["name"] not in servers:
            continue
        hostname = (
            server["wireguard_ip"]
            if wg and "wireguard_ip" in server
            else server["ssh_hostname"]
        )
        port = 22 if wg else server.get("ssh_port", 22)
        cache_key = f"{hostname}-{port}"
        if cache_key not in ssh_cache:
            key_path = path_to_config_file(server["ssh_key"]).resolve()
            if not os.path.exists(key_path):
                raise Exception(f"Can't find ssh key {key_path}")
            username = server["ssh_user"]

            try:
                connect = retry_call(
                    router.ssh,
                    exceptions=EofError,
                    tries=3,
                    fkwargs={
                        "hostname": hostname,
                        "port": port,
                        "username": username,
                        "identity_file": key_path,
                        "check_host_keys": "accept",
                        "python_path": "python3",
                        "ssh_args": ["-o", "SendEnv DUMP_COMMAND"],
                    },
                )
            except HostKeyError:
                print(
                    f"HostKeyError while trying to login to to {username}@{hostname}:{port}. Try running the following manually: ssh {username}@{hostname} -p {port} -i {key_path}"
                )
                raise
            except StreamError:
                print(
                    f"Exception while trying to login to {username}@{hostname}:{port}"
                )
                raise

            if username != "root":
                sudo = router.sudo(
                    via=connect, python_path="python3", preserve_env=True
                )
                ssh_cache[cache_key] = sudo
            else:
                ssh_cache[cache_key] = connect
        data = create_data(server=server)
        calls.append(ssh_cache[cache_key].call_async(func, data, *args, **kwargs))

    infos: List[Any] = []
    errors: List[Exception] = []
    for call in calls:
        try:
            info = call.get().unpickle()
            if info is not None:
                assert isinstance(info, Dict)
                decode(cast(Dict[Union[str, bytes], Union[str, bytes]], info))
            infos.append(info)
        except Error as e:
            print("Got error", e)
            errors.append(e)

    if len(errors) > 0:
        raise Exception(errors)

    return {"infos": infos, "data": data}


def do(
    data: Dict[str, Any], transmitmodules: TransmitModules, name: str, dry_run: bool
):
    os.environ[DRY_RUN_ENV] = str(dry_run)
    set_data(data)
    modules = makereal(transmitmodules)
    return runfunc(modules, name)


def internal_runner(
    router: Router,
    servers: list[str],
    modules: Modules,
    local_func: str,
    run_func: str,
    parse_func: str,
    dry_run: bool,
) -> None:
    for module in modules:
        runfunc([module], local_func)
        infos = main(router, servers, do, maketransmit([module]), run_func, dry_run)
        for info in infos["infos"]:
            os.environ[DRY_RUN_ENV] = str(dry_run)
            runfunc([module], parse_func, info, infos["data"])
            for module_name in info:
                for per_node in info[module_name]:
                    if not isinstance(per_node, Dict):
                        continue
                    if "selector" not in per_node:
                        continue
                    config_path = other_config_file("selectors.json")
                    if os.path.exists(config_path):
                        selector_config: Dict[str, object] = json.load(
                            open(config_path)
                        )
                    else:
                        selector_config = {}
                    merge(
                        selector_config, cast(Dict[str, object], per_node["selector"])
                    )
                    selector_config = dict(sorted(selector_config.items()))
                    set_file_contents(
                        config_path, json.dumps(selector_config, indent=2)
                    )


def generate_dependencies(modules: Modules):
    tree: Dict[str, List[Module]] = {}
    mapping: Dict[str, Module] = {}
    checked: List[Module] = []
    needs_dependencies = list(modules) + [core]
    modules = []
    while len(needs_dependencies) > 0:
        check = freeze_module(needs_dependencies.pop())
        checked.append(check)
        key = str(maketransmit_single(check))
        tree[key] = []
        mapping[key] = check
        new_dependencies = cast(
            Dict[str, List[Modules]], runfunc([check], "dependencies")
        )
        for item in new_dependencies.values():
            for new_dependency in item[0]:
                new_dependency = freeze_module(new_dependency)
                tree[key].append(new_dependency)
                if new_dependency in needs_dependencies or new_dependency in checked:
                    continue
                needs_dependencies.append(new_dependency)

    count = 0
    while len(checked) > 0:
        count += 1
        if count == 100:
            raise Exception((modules, checked))
        for key, value in tree.items():
            mapped = mapping[key]
            if mapped in modules:
                continue
            for dep in value:
                if dep not in modules:
                    break
            else:
                modules.append(mapped)
                checked.remove(mapped)

    return modules


SERVER_FILTER = Callable[[ServerDict], bool]

ALL_SERVERS: SERVER_FILTER = lambda _: True


def server_name_filter(name: str) -> SERVER_FILTER:
    return lambda server: server["name"] == name


def server_not_name_filter(name: str) -> SERVER_FILTER:
    return lambda server: server["name"] != name


def server_role_picker(name: str) -> SERVER_FILTER:
    def inner(server: ServerDict) -> bool:
        picked_server = core._index_fn(name)  # pyright: ignore[reportPrivateUsage]
        return picked_server == server

    return inner


def run(
    args: List[str],
    modules: Union[Modules, dict[Callable[[ServerDict], bool], Modules]],
):
    logging.basicConfig()
    logging.root.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(prog="paracrine")
    parser.add_argument("-i", "--inventory-path", dest="inventory_path", required=True)
    parser.add_argument("-a", "--apply", default=False, action="store_true")
    parsed_args = parser.parse_args(args)
    set_config(parsed_args.inventory_path)

    all_servers = get_config()["servers"]

    if isinstance(modules, dict):
        module_mapping = dict(
            [(k, generate_dependencies(v)) for (k, v) in modules.items()]
        )
        all_modules: Modules = []
        for modules in module_mapping.values():
            for module in modules:
                if module not in all_modules:
                    all_modules.append(module)
    else:
        all_modules = generate_dependencies(modules)
        module_mapping = {ALL_SERVERS: all_modules}

    module_descriptions = dict(
        [(module, maketransmit_single(module)) for module in all_modules]
    )

    all_server_names = set([server["name"] for server in all_servers])
    server_modules: dict[str, list[TransmitModule]] = dict(
        [(name, []) for name in all_server_names]
    )

    for server_filter, specific_modules in module_mapping.items():
        for server in all_servers:
            if not server_filter(server):
                continue
            server_modules[server["name"]].extend(
                [module_descriptions[module] for module in specific_modules]
            )

    modules_for_server: dict[TransmitModule, list[str]] = {}

    for server_name, specific_modules in server_modules.items():
        for module in specific_modules:
            if module not in modules_for_server:
                modules_for_server[module] = [server_name]
            else:
                modules_for_server[module].append(server_name)

    print("Running:")
    for module, module_name in module_descriptions.items():
        print(f"* {unfreeze_module(module_name)}", end="")
        servers = set(modules_for_server.get(module_name, []))
        if servers == all_server_names:
            print("")
        else:
            print(f" {list(servers)}")

    print("")

    def do_runs(router: Router) -> None:
        to_run_modules: list[Module] = []
        to_run_servers = []

        for module in all_modules:
            module_description = module_descriptions[module]
            if module_description not in modules_for_server:
                continue

            wanted_servers = sorted(set(modules_for_server[module_description]))
            if to_run_servers != [] and to_run_servers != wanted_servers:
                internal_runner(
                    router,
                    to_run_servers,
                    to_run_modules,
                    "local",
                    "run",
                    "parse_return",
                    not parsed_args.apply,
                )
                to_run_modules = []

            to_run_servers = wanted_servers
            to_run_modules.append(unfreeze_module(module))

        internal_runner(
            router,
            to_run_servers,
            to_run_modules,
            "local",
            "run",
            "parse_return",
            not parsed_args.apply,
        )

    run_with_router(do_runs)



from paracrine.deps import Modules, maketransmit
from paracrine.helpers.config import set_data
from paracrine.runner import generate_dependencies
from paracrine.services import cockroachdb, wireguard


def test_deps_ordering():
    modules: Modules = [wireguard, cockroachdb]
    set_data({"templates": "", "inventory": {"servers": []}})
    modules = generate_dependencies(modules)

    assert maketransmit(modules) == [
        "paracrine.runners.core",
        ("paracrine.services.cockroachdb.certs", {"versions": {}}),
        "paracrine.services.wireguard.bootstrap",
        "paracrine.services.wireguard.core",
        "paracrine.services.wireguard",
        ("paracrine.services.cockroachdb.node", {"versions": {}}),
        ("paracrine.services.cockroachdb.init", {"versions": {}}),
        "paracrine.services.cockroachdb",
    ]


def test_cockroach_versions():
    modules: Modules = [cockroachdb]
    set_data({"templates": "", "inventory": {"servers": [{"name": "foo"}]}})
    modules = generate_dependencies(modules)

    assert maketransmit(modules) == [
        "paracrine.runners.core",
        ("paracrine.services.cockroachdb.certs", {"versions": {"foo": "23.1.1"}}),
        "paracrine.services.wireguard.bootstrap",
        "paracrine.services.wireguard.core",
        "paracrine.services.wireguard",
        ("paracrine.services.cockroachdb.node", {"versions": {"foo": "23.1.1"}}),
        ("paracrine.services.cockroachdb.init", {"versions": {"foo": "23.1.1"}}),
        "paracrine.services.cockroachdb",
    ]



import json
from typing import Dict, List, cast

from ...helpers.config import other_config_file
from ...helpers.fs import set_file_contents
from .. import wireguard
from . import wipe_old_master
from .common import MASTER_FILE, get_existing_masters, get_master_ip

options = {}


def dependencies():
    return [wireguard, wipe_old_master]


def run():
    master_ip = get_master_ip()
    return {"master_ip": master_ip}


def parse_return(infos: List[Dict[str, object]]) -> None:
    local_master = infos[0]["master_ip"]
    if local_master is None:
        return
    existing_masters = set(get_existing_masters(True))
    existing_masters.add(cast(str, local_master))
    master_path = other_config_file(MASTER_FILE)
    set_file_contents(master_path, json.dumps(list(existing_masters), indent=2))



import tempfile
from pathlib import Path
from typing import BinaryIO, cast
from unittest.mock import patch

import pytest
from mitogen.ssh import HostKeyError

from paracrine.runner import run
from paracrine.services import ntp

from .conftest import set_config_data


def test_empty_servers():
    with tempfile.NamedTemporaryFile() as config_file:
        set_config_data(cast(BinaryIO, config_file), {"servers": []})
        run(["-i", config_file.name], [])


def test_bad_ssh_path():
    with tempfile.NamedTemporaryFile() as config_file:
        set_config_data(
            cast(BinaryIO, config_file),
            {
                "data_path": ".",
                "servers": [
                    {
                        "ssh_hostname": "",
                        "ssh_key": "TBD",
                        "ssh_user": "",
                        "name": "foo",
                    }
                ],
            },
        )
        with pytest.raises(Exception, match="Can't find ssh key /tmp/configs/TBD"):
            run(["-i", config_file.name], [ntp])


def test_host_key_error(capsys: pytest.CaptureFixture[str]):
    with patch("mitogen.parent.Router.ssh", side_effect=HostKeyError):
        with tempfile.TemporaryDirectory() as raw_temp_directory:
            temp_directory = Path(raw_temp_directory)
            config_file_path = temp_directory.joinpath("config.yaml")
            with config_file_path.open("wb") as config_file:
                set_config_data(
                    cast(BinaryIO, config_file),
                    {
                        "data_path": ".",
                        "servers": [
                            {
                                "name": "test",
                                "ssh_hostname": "foo",
                                "ssh_key": "key_path",
                                "ssh_user": "test_user",
                            }
                        ],
                    },
                )
            configs_folder = temp_directory.joinpath("configs")
            configs_folder.mkdir()
            key_path = configs_folder.joinpath("key_path")
            key_path.open("w").write("test")
            with pytest.raises(HostKeyError):
                run(["-i", config_file_path.as_posix()], [ntp])

            res = capsys.readouterr()
            assert (
                f"Try running the following manually: ssh test_user@foo -p 22 -i {key_path.as_posix()}"
                in res.out
            ), res.out



import logging
from pathlib import Path
from typing import Dict, Union

from ...helpers.config import host, in_docker, servers
from ...helpers.network import wireguard_ip


def cockroach_url(cockroach_version: str) -> str:
    return f"https://binaries.cockroachdb.com/cockroach-v{cockroach_version}.linux-amd64.tgz"


def binary_path(cockroach_version: str) -> str:
    return f"cockroach-v{cockroach_version}.linux-amd64/cockroach"


def cockroach_binary(cockroach_version: str) -> str:
    return f"/opt/cockroach-v{cockroach_version}.linux-amd64/{binary_path(cockroach_version)}"


cockroach_versions = {
    "23.1.1": {
        "hash": "8197562ce59d1ac4f53f67c9d277827d382db13c5e650980942bcb5e5104bb4e"
    },
    "23.2.8": {
        "hash": "4f92e2076a995e2b00181594722bd807415c1b2754154274083a7aa0f5da83a3"
    },
    "24.1.2": {
        "hash": "eff2aa8c79f3668aa81e4c123e64870343690ff1772ea1be4c06d18a161dae70"
    },
}

HOME_DIR = Path("/var/lib/cockroach")
CERTS_DIR = HOME_DIR.joinpath("certs")
USER = "cockroach"


def local_node_ip() -> str:
    if in_docker():
        return "127.0.0.1"
    else:
        wip = wireguard_ip()
        assert wip is not None
        return wip


def node_count() -> int:
    return len(servers())


def calculate_version(versions: Union[None, str, Dict[str, int]]) -> Dict[str, str]:
    if versions is None:
        versions = sorted(list(cockroach_versions.keys()))[0]

    if isinstance(versions, str):
        versions = {versions: node_count()}

    logging.info(f"Cockroach versions is {versions}")

    server_list = sorted(servers(), key=lambda s: s["name"])
    version_map: Dict[str, str] = {}

    index = 0
    for version, count in sorted(versions.items()):
        for c in range(count):
            version_map[server_list[index + c]["name"]] = version
        index += count

    assert index == len(server_list)

    logging.info(f"Cockroach version map is {version_map}")
    return version_map


def version_for_host(versions: Dict[str, str]) -> str:
    return versions[host()["name"]]



import tempfile
from typing import BinaryIO, cast
from unittest.mock import MagicMock, call, patch

from callee import InstanceOf, List, String
from mitogen.parent import Router

from paracrine.deps import Module
from paracrine.helpers import cron
from paracrine.runner import run, server_name_filter, server_role_picker
from paracrine.runners import aws, certs, core
from paracrine.services import ntp, postgresql

from .conftest import set_config_data


def get_run_sets(mock_internal_runner: MagicMock):
    sets: list[tuple[list[str], list[Module]]] = []
    for call_arg in mock_internal_runner.call_args_list:
        assert call_arg == call(
            InstanceOf(Router),
            List(of=String()),
            List(),
            "local",
            "run",
            "parse_return",
            True,
        )
        sets.append((call_arg.args[1], call_arg.args[2]))

    return sets


def test_runner_one_dep():
    with patch("paracrine.runner.internal_runner") as mock_internal_runner:
        with tempfile.NamedTemporaryFile() as config_file:
            set_config_data(
                cast(BinaryIO, config_file),
                {
                    "data_path": ".",
                    "servers": [{"name": "foo"}],
                },
            )
            run(["-i", config_file.name], [ntp])

        assert get_run_sets(mock_internal_runner) == [(["foo"], [core, ntp])]


def test_runner_dep_with_extras():
    with patch("paracrine.runner.internal_runner") as mock_internal_runner:
        with tempfile.NamedTemporaryFile() as config_file:
            set_config_data(
                cast(BinaryIO, config_file),
                {
                    "data_path": ".",
                    "servers": [{"name": "foo"}],
                },
            )
            run(["-i", config_file.name], [certs])

        assert get_run_sets(mock_internal_runner) == [
            (["foo"], [core, cron, aws, certs])
        ]


def test_runner_dep_with_multiple():
    with patch("paracrine.runner.internal_runner") as mock_internal_runner:
        with tempfile.NamedTemporaryFile() as config_file:
            set_config_data(
                cast(BinaryIO, config_file),
                {
                    "data_path": ".",
                    "servers": [{"name": "foo"}, {"name": "bar"}],
                },
            )
            run(
                ["-i", config_file.name],
                {
                    server_name_filter("foo"): [ntp],
                    server_name_filter("bar"): [postgresql],
                },
            )

        assert get_run_sets(mock_internal_runner) == [
            (["bar", "foo"], [core]),
            (["foo"], [ntp]),
            (["bar"], [postgresql]),
        ]


def test_runner_one_dep_with_config():
    with patch("paracrine.runner.internal_runner") as mock_internal_runner:
        with tempfile.NamedTemporaryFile() as config_file:
            set_config_data(
                cast(BinaryIO, config_file),
                {
                    "data_path": ".",
                    "servers": [{"name": "foo"}],
                },
            )
            run(["-i", config_file.name], [(ntp, {"foo": "bar"})])

        assert get_run_sets(mock_internal_runner) == [
            (["foo"], [core, (ntp, {"foo": "bar"})])
        ]


def test_runner_server_role_picker():
    with patch("paracrine.runner.internal_runner") as mock_internal_runner:
        with tempfile.NamedTemporaryFile() as config_file:
            set_config_data(
                cast(BinaryIO, config_file),
                {
                    "data_path": ".",
                    "servers": [{"name": "foo"}],
                },
            )
            run(
                ["-i", config_file.name],
                {
                    server_role_picker("timeserver"): [ntp],
                },
            )

        assert get_run_sets(mock_internal_runner) == [(["foo"], [core, ntp])]



from typing import Dict, List, TypedDict

from .config import Host, ServerDict, host, network_config, other_config, servers


def networks_by_interface(h: Host):
    return dict([(intf["ifname"], intf) for intf in network_config(h["name"])])


class AddrInfo(TypedDict):
    family: str
    local: str


class Network(TypedDict):
    addr_info: List[AddrInfo]


def get_ipv4(network: Network):
    addrs = network["addr_info"]
    for addr in addrs:
        if addr["family"] == "inet" and addr["local"] != "":
            return addr["local"]
    return None


def external_ip(h: ServerDict) -> str:
    return other_config(h["name"])["external_ip"]


def external_ips() -> Dict[str, object]:
    return dict([(h["name"], external_ip(h)) for h in servers()])


def wireguard_ips() -> Dict[str, str]:
    return dict(
        [(h["name"], h["wireguard_ip"]) for h in servers() if "wireguard_ip" in h]
    )


def wireguard_ip():
    return host().get("wireguard_ip")


def ips_for_network(net: Dict[str, Network]):
    return [
        x for x in [get_ipv4(intf) for (_name, intf) in net.items()] if x is not None
    ]


def ips(h: Host):
    network = networks_by_interface(h)
    return ips_for_network(network)



import importlib
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from frozendict import deepfreeze, frozendict
from mergedeep import merge

from .helpers.config import clear_return_data, get_return_data, set_data

ModuleConfig = Union[frozendict[str, object], Mapping[str, object]]
Module = Union[ModuleType, Tuple[ModuleType, ModuleConfig]]
Modules = Sequence[Module]
"""Type of modules handed to `paracrine.runner.run`"""

TransmitModule = Union[str, Tuple[str, ModuleConfig]]
TransmitModules = Sequence[TransmitModule]


def freeze_module(module: Module) -> Module:
    if isinstance(module, tuple):
        (module_type, config) = module
        if isinstance(config, frozendict):
            return module
        return (module_type, deepfreeze(config))
    else:
        return module


def undeepfreeze(fd: frozendict[str, object]) -> dict[str, object]:
    ret: dict[str, object] = {}
    for key, value in fd.items():
        if isinstance(value, frozendict):
            ret[key] = undeepfreeze(cast(frozendict[str, object], value))
        else:
            ret[key] = value

    return ret


T = TypeVar("T", bound=Union[Module, TransmitModule])


def unfreeze_module(module: T) -> T:
    if isinstance(module, tuple):
        (module_type, config) = module
        if isinstance(config, frozendict):
            return (
                module_type,
                undeepfreeze(config),
            )  # pyright: ignore[reportReturnType]

    return module


def runfunc(
    modules: Modules,
    name: str,
    arguments: Dict[str, Any] = {},
    data: Mapping[str, object] = {},
) -> Dict[str, Any]:
    ret: Dict[str, List[Mapping[str, object]]] = {}

    def run(module: ModuleType, options: ModuleConfig):
        func: Union[
            Callable[[], Optional[Dict[str, object]]],
            Callable[[str], Optional[Dict[str, object]]],
            None,
        ] = getattr(module, name, None)
        if func is not None:
            if isinstance(options, frozendict):
                options = undeepfreeze(options)
            setattr(module, "options", options)
            clear_return_data()
            if data != {}:
                set_data(data)
            try:
                if module.__name__ not in ret:
                    ret[module.__name__] = []
                if module.__name__ in arguments:
                    info = func(
                        arguments[module.__name__]  # pyright: ignore[reportCallIssue]
                    )
                else:
                    info = func()  # pyright: ignore[reportCallIssue]
                if isinstance(info, Dict):
                    info = merge({}, info, get_return_data())
                elif info is None:
                    info = get_return_data()
                ret[module.__name__].append(info)
            except Exception:
                print(f"Error while running {name} for {module.__name__}")
                raise

    for module in modules:
        if isinstance(module, ModuleType):
            run(module, {})
        else:
            run(module[0], module[1])
    return ret


def maketransmit_single(module: Module) -> TransmitModule:
    if isinstance(module, ModuleType):
        return module.__name__
    else:
        return (module[0].__name__, module[1])


def maketransmit(modules: Modules) -> TransmitModules:
    return [maketransmit_single(module) for module in modules]


def makereal(modules: TransmitModules) -> Modules:
    ret: Modules = []
    for module in modules:
        if isinstance(module, str):
            ret.append(importlib.import_module(module))
        else:
            ret.append((importlib.import_module(module[0]), module[1]))

    return ret



from typing import Dict, cast

from ...helpers.fs import run_with_marker
from ...runners.core import use_this_host
from . import certs, node
from .common import (
    CERTS_DIR,
    HOME_DIR,
    cockroach_binary,
    local_node_ip,
    version_for_host,
)

options: Dict[str, object] = {}


def dependencies():
    return [(node, options), (certs, {"versions": options["versions"]})]


def run():
    version = version_for_host(cast(Dict[str, str], options["versions"]))
    if not use_this_host("cockroach-init"):
        return
    COCKROACH_PORT = options.get("COCKROACH_PORT", 26257)
    run_with_marker(
        HOME_DIR.joinpath("init_done"),
        f"{cockroach_binary(version)} init --certs-dir={CERTS_DIR} --host={local_node_ip()}:{COCKROACH_PORT}",
        run_if_command_changed=False,
    )
